# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid.tests.testcase import PyramidDBTestCase
from anyblok.registry import RegistryManagerException


class TestCrudBlok(PyramidDBTestCase):

    blok_entry_points = ('bloks', 'test_bloks')

    def test_predicate(self):
        registry = self.init_registry(None)
        self.webserver.get('/bloks1', status=404)
        self.webserver.get('/bloks2', status=200)
        registry.upgrade(install=('test_rest_api_2',))
        self.webserver.get('/bloks1', status=200)
        self.webserver.get('/bloks2', status=200)

    def test_multi_primary_keys(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test_rest_api_2',))
        resp = self.webserver.get(
            '/column/Model.System.Model/name', status=200)
        self.assertEqual(resp.json_body, {
            'autoincrement': False,
            'code': 'system_model.name',
            'entity_type': 'Model.System.Column',
            'foreign_key': None,
            'ftype': 'String',
            'label': 'Name',
            'model': 'Model.System.Model',
            'name': 'name',
            'nullable': False,
            'primary_key': True,
            'remote_model': None,
            'unique': None
        })

    def test_bad_model(self):
        registry = self.init_registry(None)
        registry.upgrade(install=('test_rest_api_2',))
        with self.assertRaises(RegistryManagerException):
            self.webserver.get('/bad/model')


class CrudResourceBlok:

    blok_entry_points = ('bloks', 'test_bloks')

    def test_model_schema_validator_get(self):
        response = self.webserver.get(self.path % 'anyblok-core')
        self.assertEqual(response.status_code, 200)
        # self.maxDiff = None
        self.assertEqual(
            response.json_body,
            {
                'author': 'Suzanne Jean-Sébastien',
                'name': 'anyblok-core',
                'order': 0,
                'state': 'installed',
            }
        )


class TestCrudResourceBlokModelSchemaValidator(CrudResourceBlok,
                                               PyramidDBTestCase):
    """Test Customers and Addresses from
    test_bloks/test_5/views.py
    """

    def setUp(self):
        super(TestCrudResourceBlokModelSchemaValidator, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_5',))
        self.collection_path = '/bloks/v5'
        self.path = '/bloks/v5/%s'



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
# flake8: noqa



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid.tests.testcase import PyramidDBTestCase


class TestCrudServiceBase(PyramidDBTestCase):
    """Test Service from test_bloks/test_1/views.py:example_service
    This is the basic case, no validators, no schema.
    """

    blok_entry_points = ('bloks', 'test_bloks',)

    def setUp(self):
        super(TestCrudServiceBase, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_1',))

    def create_example(self, name="plop"):
        """Create a dummy example record"""
        example = self.registry.Example.insert(name=name)
        return example

    def test_example_service_get(self):
        """Example GET /anothers/{id}"""
        ex = self.create_example()
        response = self.webserver.get('/anothers/%s' % ex.id)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plop")

    def test_example_service_put(self):
        """Example PUT /anothers/{id}"""
        ex = self.create_example()
        response = self.webserver.put_json(
            '/anothers/%s' % ex.id, {'name': 'plip'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plip")

    def test_example_service_put_schema_fail_bad_path(self):
        """Example PUT schema validation fail on invalid path /anothers/{id}
        """
        fail = self.webserver.put_json(
            '/anothers/x', {'name': 'plip'}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'path')

    def test_example_service_put_schema_fail_bad_value_type(self):
        """Example PUT schema validation fail on invalid value type in `body`
        /anothers/{id}
        """
        ex = self.create_example()
        fail = self.webserver.put_json(
            '/anothers/%s' % ex.id, {'name': 0}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')

    def test_example_service_post(self):
        """Example POST /anothers/"""
        response = self.webserver.post_json('/anothers', {'name': 'plip'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plip")

    def test_example_service_post_schema_fail_bad_column_name(self):
        """Example POST schema validation fail on invalid column name /anothers/
        """
        fail = self.webserver.post_json(
            '/anothers', {'badcolumn': 'plip'}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')

    def test_example_service_post_schema_fail_bad_value_type(self):
        """Example POST schema validation fail on invalid value type /anothers/
        """
        fail = self.webserver.post_json(
            '/anothers', {'name': 0}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')


class TestCrudServiceAdvanced(PyramidDBTestCase):
    """Test Service from test_bloks/test_1/views.py:thing_service
    This is the advanced case, validators and relationnal schema
    """

    blok_entry_points = ('bloks', 'test_bloks',)

    def setUp(self):
        super(TestCrudServiceAdvanced, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_1',))

    def create_example(self, name="plop"):
        """Create a dummy example record"""
        example = self.registry.Example.insert(name=name)
        return example

    def create_examples(self, examples=None):
        """Create a dummy example collection record"""
        res = []
        if examples is None:
            examples = [dict(name="car example"),
                        dict(name="apple example"),
                        dict(name="human example")]
        for example in examples:
            res.append(self.create_example(**example))
        return res

    def create_thing(self, name="car", secret="aaa", **kwargs):
        """Create a dummy thing record"""
        if 'example_id' not in kwargs.keys():
            example = self.create_example(name="%s example" % name)
            kwargs['example_id'] = example.id
        thing = self.registry.Thing.insert(name=name, secret=secret, **kwargs)
        return thing

    def create_things(self, things=None):
        """Create a dummy thing collection record"""
        res = []
        if things is None:
            exs = self.create_examples()
            things = [dict(name="car", secret="aaa", example_id=exs[0].id),
                      dict(name="apple", secret="bbb", example_id=exs[1].id),
                      dict(name="human", secret="ccc", example_id=exs[2].id)]
        for thing in things:
            res.append(self.create_thing(**thing))
        return res

    def test_thing_service_get(self):
        """Example GET /things/{uuid}"""
        ex = self.create_thing()
        response = self.webserver.get('/things/%s' % ex.uuid)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "car")

    def test_thing_service_delete(self):
        """Example GET /things/{uuid}"""
        ex = self.create_thing()
        query = self.registry.Thing.query().filter_by(uuid=ex.uuid)
        self.assertEqual(query.count(), 1)
        response = self.webserver.delete('/things/%s' % ex.uuid)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(query.count(), 0)

    def test_thing_service_put(self):
        """Example PUT /things/{uuid}"""
        ex = self.create_thing()
        msg = dict(
            name='plip',
            secret='other',
            example_id=ex.example_id,
        )
        response = self.webserver.put_json(
            '/things/%s' % ex.uuid, msg)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plip")

    def test_thing_service_put_schema_fail_bad_path(self):
        """Example PUT schema validation fail on invalid path /things/{uuid}
        """
        fail = self.webserver.put_json(
            '/things/x', {'name': 'plip'}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertIn(
            fail.json_body.get('errors')[0].get('location'), ['path', 'body'])

    def test_thing_service_put_schema_fail_bad_value_type(self):
        """Example PUT schema validation fail on invalid value type in `body`
        /things/{uuid}
        """
        ex = self.create_thing()
        fail = self.webserver.put_json(
            '/things/%s' % ex.uuid, {'name': 0}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')

    def test_thing_service_post(self):
        """Example POST /things/"""
        ex = self.create_example()
        response = self.webserver.post_json(
            '/things',
            {'name': 'book', 'secret': 'xxx', 'example_id': ex.id}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "book")

    def test_thing_service_post_schema_fail_bad_column_name(self):
        """Example POST schema validation fail on invalid column name /things/
        """
        fail = self.webserver.post_json(
            '/things', {'badcolumn': 'plip'}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')

    def test_thing_service_post_schema_fail_bad_value_type(self):
        """Example POST schema validation fail on invalid value type /things/
        """
        fail = self.webserver.post_json(
            '/things', {'name': 0}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')

    def test_thing_collection_get(self):
        """Thing collection GET /things"""
        self.create_things()
        response = self.webserver.get('/things')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 3)
        self.assertEqual(response.json_body[0].get('name'), "car")

    def test_thing_collection_get_unknown_querystring(self):
        """Thing collection GET /things"""
        self.create_things()
        with self.assertRaises(KeyError):
            self.webserver.get('/things?foo=bar')

    def test_thing_collection_get_with_querystring(self):
        """Thing collection GET /things?querystring"""
        self.create_things()
        response = self.webserver.get('/things?filter[name][like]=car')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.headers.get('X-Total-Records')), 1)
        self.assertEqual(int(response.headers.get('X-Count-Records')), 1)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "car")

    def test_thing_collection_get_with_bad_querystring(self):
        """Thing collection GET fail on bad /things?badquerystring"""
        self.create_things()
        fail = self.webserver.get('/things?filter[name][oops]=car', status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'querystring')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck BRET <franckbret@gmail.com>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_marshmallow import SchemaWrapper
from cornice.resource import view as cornice_view
from pyramid.security import Deny, Allow, Everyone, ALL_PERMISSIONS
from pyramid.httpexceptions import HTTPUnauthorized, HTTPNotFound
from anyblok_pyramid_rest_api.querystring import QueryString
from types import MethodType
from .validator import (collection_get_validator, collection_post_validator,
                        get_validator, delete_validator, put_validator,
                        patch_validator)
from marshmallow import ValidationError
from contextlib import contextmanager
from logging import getLogger

logger = getLogger(__name__)


@contextmanager
def saved_errors_in_request(request):
    try:
        yield
    except Exception as e:
        logger.exception(e)
        request.errors.add(
            'body', '500 Internal Server Error', str(e))
        request.errors.status = 500


def get_path(request):
    """Ensure we get a valid path
    """
    if 'path' in request.validated.keys():
        return request.validated.get('path')
    else:
        return request.matchdict


def update_from_query_string(request, Model, query, adapter):
    headers = request.response.headers
    if request.params:
        # TODO: Implement schema validation to use request.validated
        querystring = QueryString(request, Model, adapter=adapter)
        total_query = querystring.update_sqlalchemy_query(
            query, only_filter=True)
        query = querystring.update_sqlalchemy_query(query)
        # TODO: Advanced pagination with Link Header
        # Link: '<https://api.github.com/user/repos?page=3&per_page=100>;
        # rel="next",
        # <https://api.github.com/user/repos?page=50&per_page=100>;
        # rel="last"'
        headers['X-Count-Records'] = str(query.count())
        headers['X-Total-Records'] = str(total_query.count())
        # TODO: Etag / timestamp / 304 if no changes
        # TODO: Cache headers
        return query
    else:
        # no querystring, returns all records (maybe we will want to apply
        # some default filters values
        headers['X-Count-Records'] = str(query.count())
        headers['X-Total-Records'] = str(query.count())
        return query


def post_item(request, Model):
    if request.errors:
        return

    if isinstance(Model, str):
        Model = request.anyblok.registry.get(Model)

    try:
        return Model.insert(**request.validated['body'])
    except Exception as e:
        request.errors.add('body', '500 Internal Server Error', str(e))
        request.errors.status = 500


def get_items(request, Model, Adapter=None):
    if request.errors:
        return

    if isinstance(Model, str):
        Model = request.anyblok.registry.get(Model)

    query = Model.query()
    query = update_from_query_string(request, Model, query, Adapter)
    if not query.count():
        return

    return query.all()


def get_item(request, Model):
    if request.errors:
        return

    if isinstance(Model, str):
        Model = request.anyblok.registry.get(Model)

    path = get_path(request)
    model_pks = Model.get_primary_keys()
    pks = {x: path[x] for x in model_pks}
    item = Model.from_primary_keys(**pks)
    if item:
        return item
    else:
        path = ', '.join(
            ['%s=%s' % (x, y)
             for x, y in request.validated.get('path', {}).items()])
        request.errors.add(
            'path', '404 not found',
            'Resource %s with %s does not exist.' % (
                Model.__registry_name__, path))
        request.errors.status = 404


def put_item(request, Model):
    """
    """
    if request.errors:
        return

    if isinstance(Model, str):
        Model = request.anyblok.registry.get(Model)

    item = get_item(request, Model)
    if item:
        body = request.validated.get('body', request.validated)
        with saved_errors_in_request(request):
            item.update(**body)

        return item


patch_item = put_item


def delete_item(request, Model):
    """
    """
    if request.errors:
        return

    if isinstance(Model, str):
        Model = request.anyblok.registry.get(Model)

    item = get_item(request, Model)
    if item:
        with saved_errors_in_request(request):
            item.delete()


# HOOK
# * deactivate some access
#   - has_collection_get: bool default True
#   - has_collection_post: bool default True
#   - has_get: bool default True
#   - has_delete: bool default True
#   - has_patch: bool default True
#   - has_put: bool default True
# * get schema
#   - default_schema: classmethod of AnyBlokMarshmallow schema, by default use
#     model to defined it
# * get serialize schema
#   - default_serialize_schema: method of AnyBlokMarshmallow schema, by
#     default use default_schema
#   - serialize_collection_get: method of AnyBlokMarshmallow schema,
#                               by default use default_serialize_schema
#   - serialize_collection_post: method of AnyBlokMarshmallow schem
#                                by default use default_serialize_schema
#   - serialize_get: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
#   - serialize_delete: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
#   - serialize_patch: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
#   - serialize_put: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
# * get serialize opts
#   - get_serialize_opts: method return dict of option to use
# * get deserialize schema
#   - default_deserialize_schema: method of AnyBlokMarshmallow schema, by
#     default use default_schema
#   - deserialize_collection_post: method of AnyBlokMarshmallow schema,
#                                  by default use default_deserialize_schema
#   - deserialize_patch: method of AnyBlokMarshmallow schema,
#                        by default use default_deserialize_schema
#   - deserialize_put: method of AnyBlokMarshmallow schema,
#                      by default use default_deserialize_schema
# * get deserialize opts
#   - get_deserialize_opts: method return dict of option to use
# * get path schema
#   - default_path_schema: method of AnyBlokMarshmallow schema, by default use
#     default_schema
#   - path_get: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
#   - path_delete: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
#   - path_patch: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
#   - path_put: method of AnyBlokMarshmallow schema, by default use
#     default_serialize_schema
# * get path opts
#   - get_path_opts: method return dict of option to use
# * update_collection_get_filter: method to improve query to filter
# * create
# * update
# * delete_entry
# * get_model_name:


class CrudResource:
    model = None
    adapter_cls = None
    cache_default_schema = True  # TODO
    has_collection_get = True
    has_collection_post = True
    has_get = True
    has_delete = True
    has_patch = True
    has_put = True

    def __init__(self, request, **kwargs):
        self.request = request
        self.registry = self.request.anyblok.registry
        self.adapter = None
        if self.adapter_cls:
            self.adapter = self.adapter_cls(
                self.registry,
                Model=self.get_model('collection_get'))
            self.adapter.load_decorators()

    @classmethod
    def get_model_name(cls, request, rest_action=None, base=None):
        return cls.model

    def model_name(self, rest_action=None):
        return self.get_model_name(
            self.request,
            rest_action=rest_action,
            base=self.request.validated
        )

    def __acl__(self):
        if not hasattr(self, 'registry'):
            raise HTTPUnauthorized("ACL have not get AnyBlok registry")

        Blok = self.registry.System.Blok
        if not Blok.is_installed('auth'):
            return [(Allow, Everyone, ALL_PERMISSIONS)]

        userid = self.request.authenticated_userid
        if userid:
            return self.registry.User.get_acl(
                userid, self.model_name(), params=dict(self.request.matchdict)
            )

        return [(Deny, Everyone, ALL_PERMISSIONS)]

    @classmethod
    def apply_validator_schema(cls, request, part, Schema, opts, base):
        if 'context' not in opts:
            opts['context'] = {}

        opts['context']['registry'] = request.anyblok.registry
        try:
            schema = Schema(**opts)
            result = schema.load(base[part])
            request.validated[part] = result
        except ValidationError as err:
            logger.exception(err)
            errors = err.messages
            for k, v in errors.items():
                request.errors.add(
                    part, 'Validation error for %s' % part,
                    ''.join(map('{}.\n'.format, v)))

    def get_model(self, rest_action):
        return self.registry.get(self.model_name(rest_action=rest_action))

    def update_collection_get_filter(self, query):
        return query

    @classmethod
    def get_value_or_call_method(cls, attribute, *args, **kwargs):
        value = None
        if hasattr(cls, attribute):
            value = getattr(cls, attribute)
            if isinstance(value, MethodType):
                value = value(*args, **kwargs)

        return value

    @classmethod
    def default_schema(cls, name, model_name):
        name = '%s.%s.%s' % (cls.__name__, name, model_name)
        return type(name, (SchemaWrapper,), {'model': model_name})

    @classmethod
    def default_serialize_schema(cls, model_name):
        return cls.get_value_or_call_method(
            'default_schema', 'Serialize', model_name)

    @classmethod
    def get_serialize_schema(cls, rest_action, model_name):
        Schema = cls.get_value_or_call_method(
            'serialize_' + rest_action, model_name)
        if Schema is None:
            Schema = cls.get_value_or_call_method(
                'default_serialize_schema', model_name)

        return Schema

    @classmethod
    def get_serialize_opts(cls, rest_action):
        opts = {}
        if rest_action == 'collection_get':
            opts['many'] = True

        return opts

    @classmethod
    def default_deserialize_schema(cls, model_name):
        return cls.get_value_or_call_method(
            'default_schema', 'Deserialize', model_name)

    @classmethod
    def get_deserialize_schema(cls, rest_action, model_name):
        Schema = cls.get_value_or_call_method(
            'deserialize_' + rest_action, model_name)
        if Schema is None:
            Schema = cls.get_value_or_call_method(
                'default_deserialize_schema', model_name)

        return Schema

    @classmethod
    def get_deserialize_opts(cls, rest_action):
        opts = {}
        if rest_action == 'patch':
            opts['partial'] = True

        return opts

    @classmethod
    def default_path_schema(cls):
        return cls.get_value_or_call_method('default_schema', 'Path', cls.model)

    @classmethod
    def get_path_schema(cls, rest_action):
        Schema = cls.get_value_or_call_method('path_' + rest_action)
        if Schema is None:
            Schema = cls.get_value_or_call_method('default_path_schema')

        return Schema

    @classmethod
    def get_path_opts(cls, rest_action):
        opts = {'context': {'only_primary_key': True}}
        return opts

    def serialize(self, rest_action, entry):
        Schema = self.get_serialize_schema(
            rest_action, self.model_name(rest_action))
        opts = self.get_serialize_opts(rest_action)
        opts['context'] = {'registry': self.registry}
        schema = Schema(**opts)
        return schema.dump(entry)

    @cornice_view(validators=(collection_get_validator,), permission="read")
    def collection_get(self):
        if not self.has_collection_get:
            raise HTTPNotFound()

        if self.request.errors:
            return

        Model = self.get_model('collection_get')
        query = self.update_collection_get_filter(Model.query())
        query = update_from_query_string(
            self.request, Model, query, self.adapter)
        if not query.count():
            return []

        return self.serialize('collection_get', query.all())

    def create(self, Model, params):
        return Model.insert(**params)

    @cornice_view(validators=(collection_post_validator,), permission="create")
    def collection_post(self):
        if not self.has_collection_post:
            raise HTTPNotFound()

        if self.request.errors:
            return

        body = self.request.validated.get('body', self.request.validated)
        item = None
        if body:
            Model = self.get_model('collection_post')
            with saved_errors_in_request(self.request):
                item = self.create(Model, params=body)

        else:
            self.request.errors.add(
                'body', '400 bad request',
                'You can not post an empty body')
            self.request.errors.status = 400

        if item:
            return self.serialize('collection_post', item)

    @cornice_view(validators=(get_validator,), permission="read")
    def get(self):
        if not self.has_get:
            raise HTTPNotFound()

        if self.request.errors:
            return

        Model = self.get_model('get')
        item = get_item(self.request, Model)
        if item:
            return self.serialize('get', item)

    def delete_entry(self, item):
        item.delete()

    @cornice_view(validators=(delete_validator,), permission="delete")
    def delete(self):
        """
        """
        if not self.has_delete:
            raise HTTPNotFound()

        if self.request.errors:
            return

        Model = self.get_model('delete')
        item = get_item(self.request, Model)
        if item:
            with saved_errors_in_request(self.request):
                self.delete_entry(item)

        return {}

    def update(self, item, params=None):
        if params is None:
            return item

        item.update(**params)

    @cornice_view(validators=(patch_validator,), permission="update")
    def patch(self):
        """
        """
        if not self.has_patch:
            raise HTTPNotFound()

        if self.request.errors:
            return

        Model = self.get_model('patch')
        item = get_item(self.request, Model)
        if item:
            body = self.request.validated.get('body', self.request.validated)
            with saved_errors_in_request(self.request):
                self.update(item, params=body)

            return self.serialize('patch', item)

    @cornice_view(validators=(put_validator,), permission="update")
    def put(self):
        """
        """
        if not self.has_put:
            raise HTTPNotFound()

        if self.request.errors:
            return

        Model = self.get_model('put')
        item = get_item(self.request, Model)
        if item:
            body = self.request.validated.get('body', self.request.validated)
            with saved_errors_in_request(self.request):
                self.update(item, params=body)

            return self.serialize('put', item)



# This file is a part of the AnyBlok / Pyramid / REST api project
#
#    Copyright (C) 2017 Franck Bret <franckbret@gmail.com>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok_pyramid.tests.testcase import PyramidDBTestCase


class TestCrudResourceBase(PyramidDBTestCase):
    """Test CrudResource class from test_bloks/test_1/views.py:ExampleResource
    This is the basic case, no validators(except the default one,
    base_validator), no schema.
    """

    blok_entry_points = ('bloks', 'test_bloks',)

    def setUp(self):
        super(TestCrudResourceBase, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_1',))

    def create_example(self, name="plop"):
        """Create a dummy example record"""
        example = self.registry.Example.insert(name=name)
        return example

    def test_example_get(self):
        """Example GET /examples/{id}"""
        ex = self.create_example()
        response = self.webserver.get('/examples/%s' % ex.id)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plop")

    def test_example_get_bad_value_in_path(self):
        """Example FAILED GET /examples/{id}"""
        self.create_example()
        fail = self.webserver.get('/examples/0', status=404)
        self.assertEqual(fail.status_code, 404)

    def test_example_collection_post(self):
        """Example POST /examples/"""
        response = self.webserver.post_json('/examples', {'name': 'plip'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plip")

    def test_example_put(self):
        """Example PUT /examples/{id}"""
        ex = self.create_example()
        response = self.webserver.put_json(
            '/examples/%s' % ex.id, {'name': 'plip'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plip")

    def test_example_put_bad_value_in_path(self):
        """Example FAILED PUT /examples/{id}"""
        self.create_example()
        fail = self.webserver.put_json(
            '/examples/0', {'name': 'plip'}, status=404)
        self.assertEqual(fail.status_code, 404)

    def test_example_delete(self):
        """Example DELETE /examples/{id}"""
        ex = self.create_example()
        response = self.webserver.delete('/examples/%s' % ex.id)
        self.assertEqual(response.status_code, 200)
        response = self.webserver.get('/examples')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body, [])

    def test_example_delete_bad_value_in_path(self):
        """Example FAILED DELETE /examples/{id}"""
        self.create_example()
        fail = self.webserver.delete('/examples/0', status=404)
        self.assertEqual(fail.status_code, 404)

    def test_example_collection_get(self):
        """Example collection GET /examples"""
        self.create_example()
        response = self.webserver.get('/examples')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "plop")

    def test_example_collection_get_order_by_asc(self):
        """Example collection GET /examples?order_by[asc]=field"""
        names = ['air', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?order_by[asc]=name')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 2)
        self.assertEqual(response.json_body[0].get('name'), "air")

    def test_example_collection_get_order_by_desc(self):
        """Example collection GET /examples?order_by[desc]=field"""
        names = ['air', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?order_by[desc]=name')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 2)
        self.assertEqual(response.json_body[0].get('name'), "zen")

    def test_example_collection_get_filter_eq(self):
        """Example collection GET /examples?filter[name][eq]=term"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?filter[name][eq]=air')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.headers.get('X-Total-Records')), 1)
        self.assertEqual(int(response.headers.get('X-Count-Records')), 1)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "air")

    def test_example_collection_get_not_filter_eq(self):
        """Example collection GET /examples?~filter[name][eq]=term"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?~filter[name][eq]=air')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.headers.get('X-Total-Records')), 4)
        self.assertEqual(int(response.headers.get('X-Count-Records')), 4)
        self.assertEqual(len(response.json_body), 4)

    def test_example_collection_get_filter_like(self):
        """Example collection GET /examples?filter[name][like]=term"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?filter[name][like]=a')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.headers.get('X-Total-Records')), 3)
        self.assertEqual(int(response.headers.get('X-Count-Records')), 3)
        self.assertEqual(len(response.json_body), 3)
        self.assertEqual(response.json_body[0].get('name'), "air")

    def test_example_collection_get_filter_ilike(self):
        """Example collection GET /examples?filter[name][ilike]=TeRm"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?filter[name][ilike]=A')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.headers.get('X-Total-Records')), 3)
        self.assertEqual(int(response.headers.get('X-Count-Records')), 3)
        self.assertEqual(len(response.json_body), 3)
        self.assertEqual(response.json_body[0].get('name'), "air")

    def test_example_collection_get_limit(self):
        """Example collection GET /examples?limit=limit"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?limit=4')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.headers.get('X-Total-Records')), 5)
        self.assertEqual(len(response.json_body), 4)
        self.assertEqual(response.json_body[0].get('name'), "air")

    def test_example_collection_get_offset(self):
        """Example collection GET /examples?offset=offset"""
        names = ['air', 'bar', 'car', 'dot', 'zen']
        for name in names:
            self.create_example(name)
        response = self.webserver.get('/examples?offset=3')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.headers.get('X-Total-Records')), 5)
        self.assertEqual(len(response.json_body), 2)
        self.assertEqual(response.json_body[0].get('name'), "dot")


class TestCrudResourceWithDefaultSchema(PyramidDBTestCase):
    """Test CrudResource class from
    test_bloks/test_1/views.py:ExampleResourceBaseValidator.

    This is the basic case, no validators(except the default one,
    base_validator), no schema.
    """

    blok_entry_points = ('bloks', 'test_bloks',)

    def setUp(self):
        super(TestCrudResourceWithDefaultSchema, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_1',))

    def create_example(self, name="plop"):
        """Create a dummy example record"""
        example = self.registry.Example.insert(name=name)
        return example

    def test_example_get(self):
        """Example GET /with/default/schema/examples/{id}"""
        ex = self.create_example()
        response = self.webserver.get(
            '/with/default/schema/examples/%s' % ex.id)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plop")

    def test_example_get_bad_value_in_path(self):
        """Example FAILED GET /with/default/schema/examples/{id}"""
        self.create_example()
        fail = self.webserver.get('/with/default/schema/examples/0', status=404)
        self.assertEqual(fail.status_code, 404)


class CrudResourceSchema:
    blok_entry_points = ('bloks', 'test_bloks',)

    def create_customer(self, name="bob", tag_name="green", zipcode="000"):
        """Create a dummy customer record"""
        tag = self.registry.Tag.insert(name=tag_name)
        customer = self.registry.Customer.insert(name=name)
        customer.tags.append(tag)
        city = self.registry.City.insert(name="nowhere", zipcode=zipcode)
        self.registry.Address.insert(
            customer=customer, city=city, street="Dead end street")
        return customer

    def test_example_collection_get(self):
        """Example collection GET /examples"""
        self.create_customer()
        response = self.webserver.get(self.collection_path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "bob")
        self.assertEqual(len(response.json_body[0].get('addresses')), 1)
        self.assertEqual(
            response.json_body[0].get('addresses')[0].get('city').get(
                'zipcode'), "000")

    def test_customer_get(self):
        """Customer GET /customers/{id}"""
        cu = self.create_customer()
        response = self.webserver.get(self.path % cu.id)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "bob")
        self.assertEqual(len(response.json_body.get('addresses')), 1)
        self.assertEqual(
            response.json_body.get('addresses')[0].get('city').get('zipcode'),
            "000")

    def test_customer_get_bad_value_in_path(self):
        """Customer FAILED GET /customers/{id}"""
        fail = self.webserver.get(self.path % '0', status=404)
        self.assertEqual(fail.status_code, 404)

    def test_customer_collection_post(self):
        """Customer POST /customers"""
        response = self.webserver.post_json(self.collection_path,
                                            {'name': 'plip'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plip")

    def test_customer_collection_post_partial(self):
        """Customer POST partial /customers"""
        response = self.webserver.post_json(self.collection_path,
                                            {'name': 'plip'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "plip")

    def test_customer_collection_post_empty_body(self):
        """Customer POST empty body /customers"""
        fail = self.webserver.post_json(
            self.collection_path, {}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('description'),
            'Missing data for required field..\n')

    def test_customer_collection_post_bad_key_in_body(self):
        """Customer POST bad key in body /customers"""
        fail = self.webserver.post_json(
            self.collection_path, {'unexistingkey': 'plip'}, status=400)
        self.assertEqual(fail.status_code, 400)
        self.assertEqual(fail.json_body.get('status'), 'error')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('location'), 'body')
        self.assertEqual(
            fail.json_body.get('errors')[0].get('name'),
            'Validation error for body')

    def test_customer_put(self):
        """Customer PUT /customers/{id}"""
        ex = self.create_customer()
        response = self.webserver.head(self.path % ex.id)  # fix headers
        response = self.webserver.put_json(
            self.path % ex.id, {'name': 'bobby'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "bobby")

    def test_customer_patch(self):
        """Customer PATCH /customers/{id}"""
        ex = self.create_customer()
        response = self.webserver.head(self.path % ex.id)  # fix headers
        response = self.webserver.patch_json(
            self.path % ex.id, {'name': 'bobby'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body.get('name'), "bobby")

    def test_customer_put_bad_value_in_path(self):
        """Customer FAILED PUT /customers/{id}"""
        self.create_customer()
        fail = self.webserver.put_json(
            self.path % '0', {'name': 'plip'}, status=404)
        self.assertEqual(fail.status_code, 404)

    def test_customer_delete(self):
        """Customer DELETE /customers/{id}"""
        ex = self.create_customer()
        response = self.webserver.delete(self.path % ex.id)
        self.assertEqual(response.status_code, 200)
        response = self.webserver.get(self.collection_path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json_body, [])

    def test_customer_delete_bad_value_in_path(self):
        """Customer FAILED DELETE /customers/{id}"""
        fail = self.webserver.get(self.path % '0', status=404)
        self.assertEqual(fail.status_code, 404)


class TestCrudResourceModelSchema(CrudResourceSchema, PyramidDBTestCase):
    """Test Customers and Addresses from
    test_bloks/test_3/views.py
    """

    def setUp(self):
        super(TestCrudResourceModelSchema, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_3',))
        self.collection_path = '/customers/v3'
        self.path = '/customers/v3/%s'


class TestCrudResourceWithAdapter(CrudResourceSchema, PyramidDBTestCase):
    """Test Customers and Addresses from
    test_bloks/test_3/views.py
    """

    def setUp(self):
        super(TestCrudResourceWithAdapter, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_4',))
        self.collection_path = '/customers/v4'
        self.path = '/customers/v4/%s'

    def create_adapter_customers(self):
        self.create_customer()
        self.create_customer(name="robert", tag_name="orange", zipcode="001")

    def test_adapter_get_all_without_tag_or_custom_filter(self):
        self.create_adapter_customers()
        response = self.webserver.get(self.collection_path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 2)

    def test_adapter_tag(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=green"
        response = self.webserver.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "bob")

    def test_adapter_unexisting_tag(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=unexisting"
        response = self.webserver.get(path, status=400)
        self.assertEqual(response.status_code, 400)

    def test_adapter_wrong_tag(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=wrong"
        response = self.webserver.get(path, status=400)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(len(response.json_body), 2)

    def test_adapter_tags_1(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tags=green"
        response = self.webserver.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "bob")

    def test_adapter_tags_2(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tags=green,orange"
        response = self.webserver.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 0)

    def test_adapter_tags_3(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tags=orange"
        response = self.webserver.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "robert")

    def test_adapter_tags_and_context_1(self):
        self.create_adapter_customers()
        path = self.collection_path + "?tag=colors&context[colors]=orange"
        response = self.webserver.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "robert")

    def test_adapter_tags_and_context_2(self):
        self.create_adapter_customers()
        path = self.collection_path
        path += "?tag=colors&context[colors]=green,orange"
        response = self.webserver.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 2)

    def test_adapter_customer_filter(self):
        self.create_adapter_customers()
        path = self.collection_path + "?filter[addresses.city][ilike]=001"
        response = self.webserver.get(path)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json_body), 1)
        self.assertEqual(response.json_body[0].get('name'), "robert")

    def test_adapter_wrong_filter(self):
        self.create_adapter_customers()
        path = self.collection_path + "?filter[addresses.other][ilike]=001"
        response = self.webserver.get(path, status=400)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(len(response.json_body), 2)


class TestCrudResourceModelSchemaValidator(CrudResourceSchema,
                                           PyramidDBTestCase):
    """Test Customers and Addresses from
    test_bloks/test_5/views.py
    """

    def setUp(self):
        super(TestCrudResourceModelSchemaValidator, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_5',))
        self.collection_path = '/customers/v5'
        self.path = '/customers/v5/%s'


class TestCrudResourceModelSchemaValidatorAndAuth(
    CrudResourceSchema, PyramidDBTestCase
):
    """Test Customers and Addresses from
    test_bloks/test_5/views.py
    """

    def setUp(self):
        super(TestCrudResourceModelSchemaValidatorAndAuth, self).setUp()
        self.registry = self.init_registry(None)
        self.registry.upgrade(install=('test_rest_api_5', 'auth-password'))
        self.collection_path = '/customers/v5'
        self.path = '/customers/v5/%s'
        self.registry.User.insert(
            login='jssuzanne', first_name='js', last_name='Suzanne')
        self.registry.User.CredentialStore.insert(
            login='jssuzanne', password='mypassword')
        self.webserver.post_json(
            '/login', {'login': 'jssuzanne', 'password': 'mypassword'},
            status=302)

    def test_customer_get_logout(self):
        """Customer GET /customers/{id} [UNLOGGED]"""
        cu = self.create_customer()
        self.webserver.post_json('/logout', {}, status=302)
        self.webserver.get(self.path % cu.id, status=403)
